##
# File: PrdCategoryStyle.py
# Date: 4-Nov-2012  John Westbrook
#
# Updates:
#  11-Feb-2013 jdw   add style identifier
#   5-Mar-2018  jdw Py2-Py3 and refactor for Python Packaging
##
"""
Standard style details for PRD and Family data categories in the
BIRD dictionary.

"""
from __future__ import absolute_import
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.01"


from mmcif_utils.style.PdbxCategoryStyleBase import PdbxCategoryStyleBase


class PrdCategoryStyle(PdbxCategoryStyleBase):
    _styleId = "BIRD_V1"

    _categoryInfo = [('citation', 'table'),
                     ('citation_author', 'table'),
                     ('pdbx_reference_molecule_family', 'table'),
                     ('pdbx_reference_molecule_list', 'table'),
                     ('pdbx_reference_molecule', 'table'),
                     ('pdbx_reference_entity_list', 'table'),
                     ('pdbx_reference_entity_nonpoly', 'table'),
                     ('pdbx_reference_entity_link', 'table'),
                     ('pdbx_reference_entity_poly_link', 'table'),
                     ('pdbx_reference_entity_poly', 'table'),
                     ('pdbx_reference_entity_poly_seq', 'table'),
                     ('pdbx_reference_entity_sequence', 'table'),
                     ('pdbx_reference_entity_src_nat', 'table'),
                     ('pdbx_reference_molecule_details', 'table'),
                     ('pdbx_reference_molecule_synonyms', 'table'),
                     ('pdbx_reference_entity_subcomponents', 'table'),
                     ('pdbx_reference_molecule_annotation', 'table'),
                     ('pdbx_reference_molecule_features', 'table'),
                     ('pdbx_reference_molecule_related_structures', 'table'),
                     ('pdbx_prd_audit', 'table'),
                     ('pdbx_family_prd_audit', 'table'),
                     ('pdbx_reference_entity_sequence_list', 'table')
                     ]
    _cDict = {
        'citation': [
            ('_citation.id', '%s', 'str', ''),
            ('_citation.title', '%s', 'str', ''),
            ('_citation.journal_abbrev', '%s', 'str', ''),
            ('_citation.journal_volume', '%s', 'str', ''),
            ('_citation.page_first', '%s', 'str', ''),
            ('_citation.page_last', '%s', 'str', ''),
            ('_citation.pdbx_database_id_DOI', '%s', 'str', ''),
            ('_citation.pdbx_database_id_PubMed', '%s', 'str', ''),
            ('_citation.year', '%s', 'str', '')
        ],
        'citation_author': [
            ('_citation_author.citation_id', '%s', 'str', ''),
            ('_citation_author.ordinal', '%s', 'str', ''),
            ('_citation_author.name', '%s', 'str', '')
        ],

        'pdbx_reference_molecule_family': [
            ('_pdbx_reference_molecule_family.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_family.name', '%s', 'str', ''),
            ('_pdbx_reference_molecule_family.release_status', '%s', 'str', ''),
            ('_pdbx_reference_molecule_family.replaces', '%s', 'str', ''),
            ('_pdbx_reference_molecule_family.replaced_by', '%s', 'str', '')
        ],
        'pdbx_reference_molecule_list': [
            ('_pdbx_reference_molecule_list.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_list.prd_id', '%s', 'str', '')
        ],
        'pdbx_reference_molecule': [
            ('_pdbx_reference_molecule.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule.formula_weight', '%s', 'str', ''),
            ('_pdbx_reference_molecule.formula', '%s', 'str', ''),
            ('_pdbx_reference_molecule.type', '%s', 'str', ''),
            ('_pdbx_reference_molecule.type_evidence_code', '%s', 'str', ''),
            ('_pdbx_reference_molecule.class', '%s', 'str', ''),
            ('_pdbx_reference_molecule.class_evidence_code', '%s', 'str', ''),
            ('_pdbx_reference_molecule.name', '%s', 'str', ''),
            ('_pdbx_reference_molecule.represent_as', '%s', 'str', ''),
            ('_pdbx_reference_molecule.chem_comp_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule.compound_details', '%s', 'str', ''),
            ('_pdbx_reference_molecule.description', '%s', 'str', ''),
            ('_pdbx_reference_molecule.representative_PDB_id_code', '%s', 'str', ''),
            ('_pdbx_reference_molecule.release_status', '%s', 'str', ''),
            ('_pdbx_reference_molecule.replaces', '%s', 'str', ''),
            ('_pdbx_reference_molecule.replaced_by', '%s', 'str', '')
        ],
        'pdbx_reference_entity_list': [
            ('_pdbx_reference_entity_list.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_list.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_list.type', '%s', 'str', ''),
            ('_pdbx_reference_entity_list.details', '%s', 'str', ''),
            ('_pdbx_reference_entity_list.component_id', '%s', 'str', '')
        ],
        'pdbx_reference_entity_nonpoly': [
            ('_pdbx_reference_entity_nonpoly.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_nonpoly.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_nonpoly.details', '%s', 'str', ''),
            ('_pdbx_reference_entity_nonpoly.name', '%s', 'str', ''),
            ('_pdbx_reference_entity_nonpoly.chem_comp_id', '%s', 'str', '')
        ],
        'pdbx_reference_entity_link': [
            ('_pdbx_reference_entity_link.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.link_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.details', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.ref_entity_id_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.ref_entity_id_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.entity_seq_num_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.entity_seq_num_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.comp_id_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.comp_id_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.atom_id_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.atom_id_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.value_order', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.component_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.component_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.nonpoly_res_num_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.nonpoly_res_num_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_link.link_class', '%s', 'str', '')
        ],
        'pdbx_reference_entity_poly_link': [
            ('_pdbx_reference_entity_poly_link.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.link_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.details', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.component_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.entity_seq_num_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.entity_seq_num_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.comp_id_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.comp_id_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.atom_id_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.atom_id_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.insert_code_1', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.insert_code_2', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_link.value_order', '%s', 'str', '')
        ],
        'pdbx_reference_entity_poly': [
            ('_pdbx_reference_entity_poly.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly.type', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly.db_code', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly.db_name', '%s', 'str', '')
        ],
        'pdbx_reference_entity_poly_seq': [
            ('_pdbx_reference_entity_poly_seq.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_seq.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_seq.mon_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_seq.parent_mon_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_seq.num', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_seq.observed', '%s', 'str', ''),
            ('_pdbx_reference_entity_poly_seq.hetero', '%s', 'str', '')
        ],
        'pdbx_reference_entity_sequence': [
            ('_pdbx_reference_entity_sequence.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence.type', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence.NRP_flag', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence.one_letter_codes', '%s', 'str', '')
        ],
        'pdbx_reference_entity_src_nat': [
            ('_pdbx_reference_entity_src_nat.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.ordinal', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.organism_scientific', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.strain', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.taxid', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.atcc', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.db_code', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.db_name', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.source', '%s', 'str', ''),
            ('_pdbx_reference_entity_src_nat.source_id', '%s', 'str', '')
        ],
        'pdbx_reference_molecule_details': [
            ('_pdbx_reference_molecule_details.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_details.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_details.ordinal', '%s', 'str', ''),
            ('_pdbx_reference_molecule_details.source', '%s', 'str', ''),
            ('_pdbx_reference_molecule_details.source_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_details.text', '%s', 'str', '')
        ],
        'pdbx_reference_molecule_synonyms': [
            ('_pdbx_reference_molecule_synonyms.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_synonyms.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_synonyms.ordinal', '%s', 'str', ''),
            ('_pdbx_reference_molecule_synonyms.name', '%s', 'str', ''),
            ('_pdbx_reference_molecule_synonyms.source', '%s', 'str', ''),
            ('_pdbx_reference_molecule_synonyms.chem_comp_id', '%s', 'str', '')
        ],
        'pdbx_reference_entity_subcomponents': [
            ('_pdbx_reference_entity_subcomponents.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_subcomponents.seq', '%s', 'str', ''),
            ('_pdbx_reference_entity_subcomponents.chem_comp_id', '%s', 'str', '')
        ],
        'pdbx_reference_molecule_annotation': [
            ('_pdbx_reference_molecule_annotation.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.ordinal', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.text', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.type', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.support', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.source', '%s', 'str', ''),
            ('_pdbx_reference_molecule_annotation.chem_comp_id', '%s', 'str', '')
        ],
        'pdbx_reference_molecule_features': [
            ('_pdbx_reference_molecule_features.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.ordinal', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.source_ordinal', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.type', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.value', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.source', '%s', 'str', ''),
            ('_pdbx_reference_molecule_features.chem_comp_id', '%s', 'str', '')
        ],
        'pdbx_reference_molecule_related_structures': [
            ('_pdbx_reference_molecule_related_structures.family_prd_id', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.ordinal', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.db_name', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.db_code', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.db_accession', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.name', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.formula', '%s', 'str', ''),
            ('_pdbx_reference_molecule_related_structures.citation_id', '%s', 'str', '')
        ],
        'pdbx_prd_audit': [
            ('_pdbx_prd_audit.prd_id', '%s', 'str', ''),
            ('_pdbx_prd_audit.date', '%s', 'str', ''),
            ('_pdbx_prd_audit.annotator', '%s', 'str', ''),
            ('_pdbx_prd_audit.processing_site', '%s', 'str', ''),
            ('_pdbx_prd_audit.details', '%s', 'str', ''),
            ('_pdbx_prd_audit.action_type', '%s', 'str', '')
        ],
        'pdbx_family_prd_audit': [
            ('_pdbx_family_prd_audit.family_prd_id', '%s', 'str', ''),
            ('_pdbx_family_prd_audit.date', '%s', 'str', ''),
            ('_pdbx_family_prd_audit.annotator', '%s', 'str', ''),
            ('_pdbx_family_prd_audit.processing_site', '%s', 'str', ''),
            ('_pdbx_family_prd_audit.details', '%s', 'str', ''),
            ('_pdbx_family_prd_audit.action_type', '%s', 'str', '')
        ],
        'pdbx_reference_entity_sequence_list': [
            ('_pdbx_reference_entity_sequence_list.prd_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence_list.ref_entity_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence_list.v_id', '%s', 'str', ''),
            ('_pdbx_reference_entity_sequence_list.three_letter_codes', '%s', 'str', '')
        ]

    }

    _excludeList = []
    #_excludeList=['_pdbx_family_prd_audit.annotator', '_pdbx_family_prd_audit.details']
    _excludeList = ['_pdbx_prd_audit.annotator', '_pdbx_prd_audit.details', '_pdbx_reference_entity_src_nat.atcc',
                    '_pdbx_reference_entity_src_nat.source_id', '_pdbx_reference_entity_sequence.one_letter_codes']
    _suppressList = []
    #

    def __init__(self):
        super(PrdCategoryStyle, self).__init__(styleId=PrdCategoryStyle._styleId,
                                               catFormatL=PrdCategoryStyle._categoryInfo,
                                               catItemD=PrdCategoryStyle._cDict,
                                               excludeList=PrdCategoryStyle._excludeList,
                                               suppressList=PrdCategoryStyle._suppressList)



# DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED## DEPRECATED
#
# File: PdbxCoreIoAdapter.py
# Date: 20-Jan-2012  John Westbrook
#
# Updates:
#
# 20-Jan-2012 jdw  Currently using core reader and python writer
#  2-Apr-2014 jdw  Add string() cast on all output data --
#  8-Mar-2018 jdw  DEPRECATED use mmcif.io.IoAdapterCore
##
"""
Adapter between PDBx IO classes (C++ Boost/Python wrappers) and Python persistence classes.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.01"


import sys
import os

import logging
logger = logging.getLogger(__name__)


from mmcif.core.mmciflib import ParseCifSimple, CifFile
from mmcif.api.PdbxContainers import DataContainer
from mmcif.api.DataCategory import DataCategory


class PdbxCoreIoAdapter(object):
    """ Adapter between PDBx IO classes (C++ Boost/Python wrappers) and Python persistence classes.
    """

    def __init__(self, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        #
        self.__inputPdbxfilePath = None
        #
        self.__containerList = []
        self.__containerNameList = []
        self.__containerTypeList = []
        #

    def setContainerList(self, containerList=None):
        try:
            self.__containerList = containerList
            self.__containNameList = []
            self.__containTypeList = []
            for container in self.__containerList:
                self.__containerNameList.append(container.getName())
                self.__containerTypeList.append(container.getType())
            return True
        except Exception as e:
            if (self.__verbose):
                logger.info("+ERROR- PdbxCoreIoAdaapter.setContainerList() initialization failed\n")
            logger.exception("Failing with %s" % str(e))
            return False

    def getContainer(self, containerName):
        idx = self.__getContainerIndex(containerName)
        if idx is not None:
            return self.__containerList[idx]
        return None

    def getContainerList(self):
        return self.__containerList

    def getContainerNameList(self):
        return self.__containerNameList

    def read(self, pdbxFilePath):
        """ Import the input Pdbx data file.
        """
        try:
            self.__inputPdbxFilePath = pdbxFilePath
            if (not os.access(pdbxFilePath, os.R_OK)):
                if (self.__verbose):
                    logger.info("+ERROR- PdbxCoreIoAdapter.read() Missing file %s\n" % pdbxFilePath)
                return False
            else:
                if (self.__debug):
                    logger.info("+PdbxCoreIoAdapter.read() reading from file path %s\n" % pdbxFilePath)

            return self.__readData(pdbxFilePath, self.__containerList)

        except Exception as e:
            if (self.__verbose):
                logger.info("+ERROR- PdbxCoreIoAdapter.read() Missing file %s\n" % pdbxFilePath)
            return False

    def __readData(self, pdbxFilePath, containerList=[]):
        """ Read all data/definition containers in the input PDBxfile.
        """
        # read file contents -
        #
        try:
            #myReader = ParseCifSimple(pdbxFilePath)
            myReader = ParseCifSimple(pdbxFilePath, verbose=self.__verbose, intCaseSense=0, maxLineLength=900, nullValue="?", parseLogFileName="")
            containerNameList = []
            containerNameList = list(myReader.GetBlockNames(containerNameList))
            if self.__debug:
                logger.info("+PdbxCoreIoAdapter.__readData() containerNameList %r\n" % repr(containerNameList))
            #
            for containerName in containerNameList:
                #
                aContainer = DataContainer(containerName)
                #
                block = myReader.GetBlock(containerName)
                tableNameList = []
                tableNameList = list(block.GetTableNames(tableNameList))

                for tableName in tableNameList:
                    table = block.GetTable(tableName)
                    attributeNameList = list(table.GetColumnNames())
                    if self.__debug:
                        logger.info("+PdbxCoreIoAdapter.__readData() Attribute name list %r\n" % repr(attributeNameList))
                    numRows = table.GetNumRows()
                    rowList = []
                    for iRow in range(0, numRows):
                        row = table.GetRow(iRow)
                        rowList.append(list(row))
                    aCategory = DataCategory(tableName, attributeNameList, rowList)
                    aContainer.append(aCategory)
                    #
                    if self.__debug:
                        logger.info("+PdbxCoreIoAdapter.__read() read %s length %d %d\n" % (tableName, numRows, len(rowList)))
                        # aCategory.printIt()
                containerList.append(aContainer)
            #
            for container in containerList:
                self.__containerNameList.append(container.getName())
                self.__containerTypeList.append(container.getType())
            #
            return True
        except Exception as e:
            if (self.__verbose):
                logger.info("+ERROR- PdbxCoreIoAdapter.__read() Read failed for file %s\n" % pdbxFilePath)
            logger.exception("Failing with %s" % str(e))
            return False

    def write(self, pdbxFilePath, maxLineLength=2048):
        """ Export the current containerlist to PDBx format file in the path 'pdbxFilePath'.
        """
        try:
            if self.__debug:
                logger.info("+PdbxCoreIoAdapter.write() container length %d\n" % len(self.__containerList))
            #cF = CifFile(maxLineLength=maxLineLength)
            cF = CifFile(True, self.__verbose, 0, maxLineLength, '?')
            for container in self.__containerList:
                containerName = container.getName()
                if self.__debug:
                    logger.info("+PdbxCoreIoAdapter.write() write container %s\n" % containerName)
                cF.AddBlock(containerName)
                block = cF.GetBlock(containerName)
                objNameList = container.getObjNameList()
                if self.__debug:
                    logger.info("+PdbxCoreIoAdapter.write() category length %d\n" % len(objNameList))
                #
                for objName in objNameList:
                    name, attributeNameList, rowList = container.getObj(objName).get()
                    table = block.AddTable(name)
                    for attributeName in attributeNameList:
                        table.AddColumn(attributeName)
                    for ii, row in enumerate(rowList):
                        table.AddRow()
                        table.FillRow(ii, [str(col) if col is not None else '?' for col in row])
                    block.WriteTable(table)
            cF.Write(pdbxFilePath)

        except Exception as e:
            if (self.__verbose):
                logger.info("+ERROR- PdbxCoreIoAdapter.write() write failed for file %s\n" % pdbxFilePath)
            logger.exception("Failing with %s" % str(e))
            return False

    def writeContainerList(self, pdbxFilePath, containerNameList):
        """ Export the input containerlist to PDBx format file in the path 'pdbxFilePath'.
        """
        try:
            #cF = CifFile(maxLineLength=2048)
            cF = CifFile(True, self.__verbose, 0, maxLineLength, '?')
            for containerName in containerNameList:
                idx = self.__getContainerIndex(containerName)
                if idx is not None:
                    cF.AddBlock(containerName)
                    block = cF.GetBlock(containerName)
                    objNameList = container.getObjNameList()
                    #
                    for objName in objNameList:
                        name, attributeNameList, rowList = container.getObj(objName).get()
                        table = block.AddTable(name)
                        for attributeName in attributeNameList:
                            table.AddColumn(attributeName)
                        for ii, row in enumerate(rowList):
                            table.AddRow()
                            table.FillRow(ii, [str(col) if col is not None else '?' for col in row])
                        block.WriteTable(table)
            cF.Write(pdbxFilePath)
        except Exception as e:
            if (self.__verbose):
                logger.info("+ERROR- PdbxCoreIoAdapter.write() write failed for file %s\n" % pdbxFilePath)
            logger.exception("Failing with %s" % str(e))
            return False

    def __getContainerIndex(self, name):
        try:
            return self.__containerNameList.index(name)
        except Exception as e:
            return None
    # -----------------------------------

    def __attributePart(self, name):
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1:]



##
# File: DictionaryMapper.py
# Date: 22-Jun-2015  John Westbrook
#
# Updates:
#  24-Jun-2015  jdw add parent self mapping
##
"""
Accessors for item-level mapping information between dictionary versions.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.01"

import sys
import logging
logger = logging.getLogger(__name__)

from mmcif.api.PdbxContainers import CifName


class DictionaryMapper(object):

    """ Methods for accessing category and item-level mapping information --- .
    """

    def __init__(self, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        # for each source and destination category the list of corresponding/mapped categories
        self.__srcCatMap = {}
        self.__dstCatMap = {}
        # for each source and destination category the list of items declared
        self.__srcCatItems = {}
        self.__dstCatItems = {}
        #
        # for each source and destination category the list of key items in a mapped key relation
        self.__srcCatMappedKeyItems = {}
        self.__dstCatMappedKeyItems = {}
        #
        # Item correspondences - for each source category as a dictionary of d[srcItem] = dstItem
        #
        self.__srcCatMappedItems = {}
        #
        # Item correspondences - for each destination category as a dictionary of d[dstItem] = srcItem
        #
        self.__dstCatMappedItems = {}
        #
        #  Source self mapping:   target = src_to_src and type = as_parent
        #
        self.__srcCatParentMappedItems = {}

        #  Destination self mapping:   target = dst_to_dst and type = as_parent
        #
        self.__dstCatParentMappedItems = {}

    def set(self, attrDictList):
        """  Input mapping category (pdbx_dict_item_mapping) data as an attribute dictionary list.
        """
        self.__itemMapCatObj = attrDictList
        self.__build()

    def __build(self):
        """  Create internal data structures supporting the methods in this class -

             Supported map_target/type  -- SRC_TO_DST (AS_KEY,AS_ITEM)
                                           SRC_TO_SRC (FROM_PARENT)
                                           DST_TO_DST (FROM_PARENT)

        """
        for row in self.__itemMapCatObj:
            srcItem = row['item_name_src']
            dstItem = row['item_name_dst']
            if srcItem is None or len(srcItem) == 0 or srcItem in ['.', '?'] or dstItem is None or len(dstItem) == 0 or dstItem in ['.', '?']:
                logger.debug("+DictionaryMapper.__build() - incomplete mapping for %s %s" % (srcItem, dstItem))
                continue
            srcCat = CifName.categoryPart(srcItem)
            dstCat = CifName.categoryPart(dstItem)
            mapType = str(row['map_type']).upper()
            mapTarget = str(row['map_target']).upper()

            #
            #  Special cases for limited self mapping -
            #
            if (mapType == 'FROM_PARENT' and mapTarget == 'DST_TO_DST'):
                if dstCat not in self.__dstCatItems:
                    self.__dstCatItems[dstCat] = []
                if dstItem not in self.__dstCatItems[dstCat]:
                    self.__dstCatItems[dstCat].append(dstItem)

                if dstCat not in self.__dstCatParentMappedItems:
                    self.__dstCatParentMappedItems[dstCat] = {}
                self.__dstCatParentMappedItems[dstCat][dstItem] = srcItem
                continue

            if (mapType == 'FROM_PARENT' and mapTarget == 'SRC_TO_SRC'):

                if srcCat not in self.__srcCatItems:
                    self.__srcCatItems[srcCat] = []
                if srcItem not in self.__srcCatItems[srcCat]:
                    self.__srcCatItems[srcCat].append(srcItem)
                if srcCat not in self.__srcCatParentMappedItems:
                    self.__srcCatParentMappedItems[srcCat] = {}
                self.__srcCatParentMappedItems[srcCat][srcItem] = dstItem
                continue
            #
            if mapTarget not in ['SRC_TO_DST'] and mapType not in ['AS_ITEM', 'AS_KEY']:
                logger.warning("Unsupported mapping for %s %s %s %s" % (srcItem, dstItem, mapType, mapTarget))
                continue
            #
            # For other supported mapping types --
            #
            if srcCat not in self.__srcCatMap:
                self.__srcCatMap[srcCat] = []
            if dstCat not in self.__srcCatMap[srcCat]:
                self.__srcCatMap[srcCat].append(dstCat)

            if srcCat not in self.__srcCatItems:
                self.__srcCatItems[srcCat] = []
            if srcItem not in self.__srcCatItems[srcCat]:
                self.__srcCatItems[srcCat].append(srcItem)

            if dstCat not in self.__dstCatMap:
                self.__dstCatMap[dstCat] = []
            if srcCat not in self.__dstCatMap[dstCat]:
                self.__dstCatMap[dstCat].append(srcCat)
            #
            if dstCat not in self.__dstCatItems:
                self.__dstCatItems[dstCat] = []
            if dstItem not in self.__dstCatItems[dstCat]:
                self.__dstCatItems[dstCat].append(dstItem)
            #
            #
            if mapType == 'AS_KEY':
                if (srcCat, dstCat) not in self.__srcCatMappedKeyItems:
                    self.__srcCatMappedKeyItems[(srcCat, dstCat)] = []
                self.__srcCatMappedKeyItems[(srcCat, dstCat)].append(srcItem)

                if (dstCat, srcCat) not in self.__dstCatMappedKeyItems:
                    self.__dstCatMappedKeyItems[(dstCat, srcCat)] = []
                self.__dstCatMappedKeyItems[(dstCat, srcCat)].append(dstItem)

            if (srcCat, dstCat) not in self.__srcCatMappedItems:
                self.__srcCatMappedItems[(srcCat, dstCat)] = {}
            self.__srcCatMappedItems[(srcCat, dstCat)][srcItem] = dstItem

            if (dstCat, srcCat) not in self.__dstCatMappedItems:
                self.__dstCatMappedItems[(dstCat, srcCat)] = {}
            self.__dstCatMappedItems[(dstCat, srcCat)][dstItem] = srcItem
            #

    def getSrcMappedParentItems(self, srcCatName):
        if srcCatName in self.__srcCatParentMappedItems:
            return self.__srcCatParentMappedItems[srcCatName]
        else:
            return {}

    def getDstMappedParentItems(self, dstCatName):
        if dstCatName in self.__dstCatParentMappedItems:
            return self.__dstCatParentMappedItems[dstCatName]
        else:
            return {}

    def getSrcCategoryList(self):
        """ Return the list of declared source categories.
        """
        return sorted(self.__srcCatMap.keys())

    def getSrcCategoryItemList(self, srcCatName):
        """ Return the list of declared items in the input source category.
        """
        if srcCatName in self.__srcCatItems:
            return self.__srcCatItems[srcCatName]
        else:
            return []

    def getSrcCategoryKeyItemList(self, srcCatName, dstCatName):
        """ Return the list of mapped key items in the input source category.
        """
        if (srcCatName, dstCatName) in self.__srcCatMappedKeyItems:
            return self.__srcCatMappedKeyItems[(srcCatName, dstCatName)]
        else:
            return []

    def getDstCategoryList(self):
        """ Return the list of declared destination categories.
        """
        return sorted(self.__dstCatMap.keys())

    def getDstCategoryItemList(self, dstCatName):
        """ Return the list of declared items in the input destination category.
        """
        if dstCatName in self.__dstCatItems:
            return self.__dstCatItems[dstCatName]
        else:
            return []

    def getDstCategoryKeyItemList(self, dstCatName, srcCatName):
        """ Return the list of mapped key items in the input destination category.
        """
        if (dstCatName, srcCatName) in self.__dstCatMappedKeyItems:
            return self.__dstCatMappedKeyItems[(dstCatName, srcCatName)]
        else:
            return []

    def getMappedCategoryListForSrc(self, srcCatName):
        """  Return the list of destination categories mapped to the input source category.
        """
        if srcCatName in self.__srcCatMap:
            return self.__srcCatMap[srcCatName]
        else:
            return []

    def getMappedItemsForSrc(self, srcCatName, dstCatName):
        """  Return the dictionary of mapped items in the input source category (d[srcItem]=dstItem)
        """
        if (srcCatName, dstCatName) in self.__srcCatMappedItems:
            return self.__srcCatMappedItems[(srcCatName, dstCatName)]
        else:
            return {}

    def getMappedCategoryListForDst(self, dstCatName):
        """  Return the list of source categories mapped to the input destination category.
        """
        if dstCatName in self.__dstCatMap:
            return self.__dstCatMap[dstCatName]
        else:
            return []

    def getMappedItemsForDst(self, dstCatName, srcCatName):
        """  Return the dictionary of mapped items in the input destination category (d[dstItem]=srcItem)
        """
        if (dstCatName, srcCatName) in self.__dstCatMappedItems:
            return self.__dstCatMappedItems[(dstCatName, srcCatName)]
        else:
            return {}



##
# File: PdbxDictionaryInfo.py
# Date: 19-Mar-2012  John Westbrook
#
# Updates:
#  20-Mar-2012  jdw add storage class.
#  21-Mar-2012  jdw Add method to support attribute keys in dictionary object
#  28-Jul-2012  jdw Add display view support -
#   1-Aug-2012  jdw Change initialiation in __buildView()
#   1-Aug-2012  rps Updated initialization in __buildView() to also capture displayName when encountering first entry for a category
#   2-Aug-2012  jdw Add category level view details and supporting methods.
#   6-Aug-2012  jdw Add category display name and readonly flag to item level category
#                       Attribute level methods renamed to item level methods -
#                       Make the display category name the unique identifier rather than category name.
#   7-Aug-2012  rps Updated to allow more granular handling of ITEM_LIST and ITEM_DICT relative to a given cif category
#   2-Sep-2012  jdw Add alternate metadata options.
#  21-Sep-2012  jdw Store the unique list of column names in the persisted object.
#   5-Feb-2013  jdw extend view metadata to represent multiple displayable categories within a category group.
#   7-Feb-2013  jdw simplified category organization and changed prototypes of api methods.
#  19-Feb-2017  ep  add setFromViewContainer() to loading container definition
#   8-Mar-2018  jdw using latest mmcif IO modules
#   8-Mar-2018  jdw Py2-Py3 and refactor for Python Packaging
##
"""
Repackage PDBx dictionary and view metadata by category.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.01"


import sys
import os
import shelve
import shutil
import pickle

import logging
logger = logging.getLogger(__name__)

#from mmcif_utils.persist.PdbxPyIoAdapter import PdbxPyIoAdapter as PdbxIoAdapter
from mmcif.io.IoAdapterPy import IoAdapterPy as IoAdapter
from mmcif.api.DictionaryApi import DictionaryApi
from mmcif_utils.persist.LockFile import LockFile


class PdbxDictionaryViewInfo(object):
    """ Repackages PDBx dictionary view metadata and provide accessors for this information.

    """

    def __init__(self, viewPath="pdbx_display_view_info.cif", verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        #
        self.__pathViewFile = viewPath
        #
        self.__vD = {}
        self.__setup(viewFilePath=self.__pathViewFile)

    def __setup(self, viewFilePath=None):
        """  Create the internal view object data structure from the input view data file.
        """
        if (viewFilePath is not None and os.access(viewFilePath, os.F_OK)):
            myReader = IoAdapter(self.__verbose, self.__lfh)
            containerList = myReader.readFile(inputFilePath=viewFilePath)
            viewContainer = None
            for c in containerList:
                if c.getName() == "display_view":
                    viewContainer = c
                    break
            #
            #viewContainer = myReader.getContainer(containerName="display_view")
            if viewContainer is not None:
                self.__vD = self.__buildView(viewContainer=viewContainer)
                return True
            else:
                return False
        #
        return False

    def __buildView(self, viewContainer=None):
        """  Create the internal view object data structure from the data in the input category container.

             View Data assumptions -

             within category 'pdbx_display_view_category_group_info' the combination of viewId, menu display name and group display name
                                are unique.

             within category 'pdbx_display_view_category_info' the combination of viewId, group display name, and category name
                                are unique.


             within category 'pdbx_display_view_item_info' the combination of viewId, category group display name,
                             and item name must be unique.

        """
        vD = {}
        if viewContainer is None:
            return False
        vCat = viewContainer.getObj("pdbx_display_view_item_info")
        if (vCat is None):
            return False
        if (vCat.hasAttribute("view_id") and vCat.hasAttribute("item_name") and
                vCat.hasAttribute("item_display_name") and vCat.hasAttribute("category_display_name") and
                vCat.hasAttribute("read_only_flag")):
            idView = vCat.getIndex("view_id")
            idItem = vCat.getIndex("item_name")
            idDisplay = vCat.getIndex("item_display_name")
            idCatDisplay = vCat.getIndex("category_display_name")
            idReadOnly = vCat.getIndex("read_only_flag")

            for row in vCat.getRowList():
                viewName = row[idView]
                readOnly = row[idReadOnly]
                catDisplayName = row[idCatDisplay]
                itemName = row[idItem]
                catName = self.__categoryPart(itemName)
                displayName = row[idDisplay]
                if (self.__debug):
                    logger.debug(" %s %s %s\n" % (viewName, itemName, displayName))
                if viewName not in vD:
                    vObj = {}
                    vObj["DISPLAY_CATEGORY_LIST"] = []
                    vObj["DISPLAY_CATEGORY_DICT"] = {}
                    vObj["DISPLAY_MENU_DICT"] = {}
                    vObj["DISPLAY_MENU_LIST"] = []
                    vObj["CATEGORY_GROUP_DICT"] = {}
                    vD[viewName] = vObj
                #
                if catDisplayName not in vD[viewName]["DISPLAY_CATEGORY_DICT"]:
                    vD[viewName]["DISPLAY_CATEGORY_LIST"].append(catDisplayName)
                    vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName] = {}

                if catName not in vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName]:
                    vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName] = {}
                    vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_LIST'] = []
                    vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_DICT'] = {}

                vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_LIST'].append(itemName)

                if itemName not in vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_DICT']:
                    vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_DICT'][itemName] = {}

                vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_DICT'][itemName]['DISPLAY_NAME'] = displayName
                vD[viewName]["DISPLAY_CATEGORY_DICT"][catDisplayName][catName]['ITEM_DICT'][itemName]['READ_ONLY'] = readOnly

        #
        vCat = viewContainer.getObj("pdbx_display_view_category_group_info")
        if ((vCat is not None) and (vCat.hasAttribute("view_id") and
                                    vCat.hasAttribute("category_group_display_name") and
                                    vCat.hasAttribute("category_menu_display_name"))):
            idView = vCat.getIndex("view_id")
            idMenuName = vCat.getIndex("category_menu_display_name")
            idGroupName = vCat.getIndex("category_group_display_name")
            for row in vCat.getRowList():
                viewName = row[idView]
                groupName = row[idGroupName]
                menuName = row[idMenuName]

                if (self.__debug):
                    logger.debug(" %s %s %s\n" % (viewName, menuName, groupName))
                if viewName not in vD:
                    vObj = {}
                    vObj["DISPLAY_CATEGORY_LIST"] = []
                    vObj["DISPLAY_CATEGORY_DICT"] = {}
                    vObj["DISPLAY_MENU_DICT"] = {}
                    vObj["DISPLAY_MENU_LIST"] = []
                    vObj["CATEGORY_GROUP_DICT"] = {}
                    vD[viewName] = vObj
                #
                if menuName not in vD[viewName]["DISPLAY_MENU_DICT"]:
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName] = {}
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_LIST'] = []
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_LIST'].append(groupName)
                    vD[viewName]["DISPLAY_MENU_LIST"].append(menuName)
                else:
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_LIST'].append(groupName)

        vCat = viewContainer.getObj("pdbx_display_view_category_info")
        if ((vCat is not None) and (vCat.hasAttribute("view_id") and
                                    vCat.hasAttribute("category_group_display_name") and
                                    vCat.hasAttribute("category_group_display_name") and
                                    vCat.hasAttribute("category_display_name") and
                                    vCat.hasAttribute("category_name") and
                                    vCat.hasAttribute("category_cardinality"))):
            idView = vCat.getIndex("view_id")
            idMenuName = vCat.getIndex("category_menu_display_name")
            idCategory = vCat.getIndex("category_name")
            idCard = vCat.getIndex("category_cardinality")
            idCatGroupName = vCat.getIndex("category_group_display_name")
            idCatDisplayName = vCat.getIndex("category_display_name")
            for row in vCat.getRowList():
                viewName = row[idView]
                menuName = row[idMenuName]
                groupName = row[idCatGroupName]
                displayName = row[idCatDisplayName]
                catName = row[idCategory]
                cardinality = row[idCard]

                if (self.__debug):
                    logger.debug(" %s %s %s %s %s\n" % (viewName, menuName, groupName, catName, cardinality))

                if viewName not in vD:
                    vObj = {}
                    vObj["DISPLAY_CATEGORY_LIST"] = []
                    vObj["DISPLAY_CATEGORY_DICT"] = {}
                    #
                    vObj["DISPLAY_MENU_LIST"] = []
                    vObj["DISPLAY_MENU_DICT"] = {}
                    vObj["CATEGORY_GROUP_DICT"] = {}
                    vD[viewName] = vObj
                #

                if menuName not in vD[viewName]["DISPLAY_MENU_DICT"]:
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName] = {}
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT'] = {}
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_LIST'] = []
                    vD[viewName]["DISPLAY_MENU_LIST"].append(menuName)

                vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_LIST'].append(groupName)

                if groupName not in vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT']:
                    vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT'][groupName] = []

                vD[viewName]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT'][groupName].append((menuName, displayName, catName, cardinality))

        return vD

    def get(self):
        """  Return the full view object.
        """
        return self.__vD

    def set(self, viewObj=None):
        """  Initialize the internal data structure with the input view object.
        """
        if (viewObj is None):
            self.__vD = {}
        else:
            self.__vD = viewObj

    def setFromViewContainer(self, viewContainer=None):
        """  Initialize the internal data structure with a viewContainer
        """
        if (viewContainer is None):
            self.__vD = {}
        else:
            self.__vD = self.__buildView(viewContainer=viewContainer)

    def getViewList(self):
        """ Return the list of defined views.
        """
        return list(self.__vD.keys())

    def getDisplayCategoryList(self, viewId=None):
        """ Return the category display names list in the input view.
        """
        if (viewId is not None and viewId in self.__vD):
            return self.__vD[viewId]['DISPLAY_CATEGORY_LIST']
        else:
            return []

    def getDisplayMenuList(self, viewId):
        """ Return the list menu display items in the input view.
        """
        if (viewId is not None and viewId in self.__vD):
            return self.__vD[viewId]['DISPLAY_MENU_LIST']
        else:
            return []

    def getCategoryGroupList(self, viewId, menuName):
        """ Return the list of category group names within the input menu of the input view.
        """
        try:
            return self.__vD[viewId]['DISPLAY_MENU_DICT'][menuName]['DISPLAY_CATEGORY_GROUP_LIST']
        except Exception as e:
            return []

    def getDisplayCategoryListInGroup(self, viewId, menuName, groupName):
        """ Return the list of category group names in the input menu in the input view.
        """
        try:
            return [t[1] for t in self.__vD[viewId]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT'][groupName]]
        except Exception as e:
            return []
    #

    def getCategoryGroupListInMenu(self, viewId, menuName):
        """ Return the display category group list in the input display menu view item..
        """
        try:
            return self.__vD[viewId]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_LIST']
        except Exception as e:
            return []

    def getCategoryName(self, viewId, menuName, categoryGroupName, categoryDisplayName):
        """ Return the category name in the input view/menu/category display name.
        """
        try:
            ret = None
            for menuName, displayName, catName, card in self.__vD[viewId]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT'][categoryGroupName]:
                if displayName == categoryDisplayName:
                    ret = catName
            return ret
        except Exception as e:
            return None

    def getCategoryCardinality(self, viewId, menuName, categoryGroupName, categoryDisplayName):
        """ Return the category name in the input view/menu/category display name.
        """
        try:
            ret = None
            for menuName, displayName, catName, card in self.__vD[viewId]["DISPLAY_MENU_DICT"][menuName]['DISPLAY_CATEGORY_GROUP_DICT'][categoryGroupName]:
                if displayName == categoryDisplayName:
                    ret = card
            return ret
        except Exception as e:
            return None

    def getItemList(self, viewId, categoryDisplayName, categoryName):
        """ Return the item list in the input category in the input view.
        """
        if (viewId in self.__vD and categoryDisplayName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'] and
                categoryName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName]):
            return self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName][categoryName]['ITEM_LIST']
        else:
            return []

    def getItemDisplayName(self, viewId, categoryDisplayName, categoryName, itemName):
        """ Return the item display name for the item in the input display category in the input view.
        """
        if (viewId in self.__vD and categoryDisplayName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'] and
                categoryName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName] and
                itemName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName][categoryName]['ITEM_DICT']):
            return self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName][categoryName]['ITEM_DICT'][itemName]['DISPLAY_NAME']
        else:
            return itemName

    def getItemReadOnlyFlag(self, viewId, categoryDisplayName, categoryName, itemName):
        """ Return the read only flag for the item in the input display category in the input view.
        """
        if (viewId in self.__vD and categoryDisplayName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'] and
                categoryName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName] and
                itemName in self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName][categoryName]['ITEM_DICT']):
            return self.__vD[viewId]['DISPLAY_CATEGORY_DICT'][categoryDisplayName][categoryName]['ITEM_DICT'][itemName]['READ_ONLY']
        else:
            return itemName

    def __attributePart(self, name):
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1:]

    def __categoryPart(self, name):
        tname = ""
        if name.startswith("_"):
            tname = name[1:]
        else:
            tname = name

        i = tname.find(".")
        if i == -1:
            return tname
        else:
            return tname[:i]


class PdbxDictionaryInfo(object):
    """ Repackages PDBx dictionary metadata by category.

    """

    def __init__(self, dictPath="mmcif_pdbx_v40.dic", verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        #
        self.__pathPdbxDict = dictPath

        # obsolete
        self.__containerList = []
        self.__containerNameList = []
        #
        self.__categoryNameList = []
        self.__categoryIndexD = {}
        #
        # Parameters to tune lock file management --
        self.__timeoutSeconds = 10
        self.__retrySeconds = 0.2
        # placeholder for LockFile object for open/close methods
        self.__lockObj = None
        self.__isExpired = True
        self.__db = None
        #
        self.__setup()

    def __setup(self):
        """ Read text dictionary and create internal lists of category names and objects.
        """
        self.__viewContainer = None
        myReader = IoAdapter(self.__verbose, self.__lfh)
        #ok = myReader.readFile(inputFilePath=self.__pathPdbxDict)
        #self.__containerNameList = myReader.getContainerNameList()
        #
        self.__containerList = myReader.readFile(inputFilePath=self.__pathPdbxDict)
        self.__containerNameList = [c.getName() for c in self.__containerList]
        self.__dApi = DictionaryApi(self.__containerList)
        self.__categoryNameList = self.__dApi.getCategoryList()
        self.__categoryIndexD = self.__dApi.getCategoryIndex()
        self.__parentD = self.__dApi.getParentDictionary()
        #
        if (self.__debug):
            self.dbg()

    def dbg(self, fName="dict.dump"):
        ofh = open(fName, 'w')
        self.__dApi.dumpFeatures(ofh)
        ofh.close()

    def assembleByIndex(self):
        """ Repackage dictionary data for each category.  Attribute data is stored in
            dictionaries using attribute index keys.
        """
        dInfo = {}
        for categoryName in self.__categoryNameList:
            self.__setCategory(categoryName)
            dI = {}
            dI['DISPLAY_NAME'] = categoryName.lower()
            dI['COLUMN_NAMES'] = self.__currentAttributeNameList
            dI['COLUMN_DISPLAY_NAMES'] = self.__currentAttributeNameList
            dI['PRIMARY_KEYS'] = self.__getKeyIndexList()
            dI['MANDATORY_COLUMNS'] = self.__getMandatoryIndexList()
            dI['MANDATORY_COLUMNS_ALT'] = self.__getAltMandatoryIndexList()
            dI['COLUMN_DISPLAY_ORDER'] = self.__getOrderIndexList(keyList=dI['PRIMARY_KEYS'], mandatoryList=dI['MANDATORY_COLUMNS'])
            dI['COLUMN_TYPES'] = self.__getTypeIndexD()
            dI['COLUMN_TYPES_ALT'] = self.__getAltTypeIndexD()
            dI['COLUMN_ENUMS'] = self.__getEnumIndexD()
            dI['COLUMN_ENUMS_ALT'] = self.__getAltEnumIndexD()
            dI['COLUMN_DESCRIPTIONS'] = self.__getDescriptionIndexD()
            dI['COLUMN_DESCRIPTIONS_ALT'] = self.__getAltDescriptionIndexD()
            dI['COLUMN_EXAMPLES'] = self.__getExampleIndexD()
            dI['COLUMN_EXAMPLES_ALT'] = self.__getAltExampleIndexD()
            dI['COLUMN_REGEX'] = self.__getRegexIndexD()
            dI['COLUMN_REGEX_ALT'] = self.__getAltRegexIndexD()
            dI['COLUMN_BOUNDARY_VALUES'] = self.__getBoundaryIndexD()
            dI['COLUMN_BOUNDARY_VALUES_ALT'] = self.__getAltBoundaryIndexD()
            dI['COLUMN_DEFAULT_VALUES'] = self.__getDefaultIndexD()
            #
            if (self.__debug):
                logger.debug("PdbxDictionaryInfo(assemble): category %s attributes %r\n" % (categoryName, self.__currentAttributeNameList))
                logger.debug("PdbxDictionaryInfo(assemble): category %s key index list %r\n" % (categoryName, dI['PRIMARY_KEYS']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s mandatory index list %r\n" % (categoryName, dI['MANDATORY_COLUMNS']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s mandatory index list (alt) %r\n" % (categoryName, dI['MANDATORY_COLUMNS_ALT']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s order index list %r\n" % (categoryName, dI['COLUMN_DISPLAY_ORDER']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type index dictionary %r\n" % (categoryName, dI['COLUMN_TYPES'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type index dictionary (alt) %r\n" % (categoryName, dI['COLUMN_TYPES_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type enum dictionary %r\n" % (categoryName, dI['COLUMN_ENUMS'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type enum dictionary (alt)  %r\n" % (categoryName, dI['COLUMN_ENUMS_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type description dictionary %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type description dictionary (alt) %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type example dictionary %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type example dictionary (alt) %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type regex dictionary %r\n" % (categoryName, dI['COLUMN_REGEX'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type regex dictionary (alt) %r\n" % (categoryName, dI['COLUMN_REGEX_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s boundary dictionary %r\n" % (categoryName, dI['COLUMN_BOUNDARY_VALUES'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s boundary dictionary (alt)  %r\n" % (categoryName, dI['COLUMN_BOUNDARY_VALUES_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s default value dictionary %r\n" % (categoryName, dI['COLUMN_BOUNDARY_VALUES'].items()))
            dInfo[categoryName] = dI
        #
        return dInfo

    def assembleByAttribute(self):
        """ Repackage dictionary data for each category.  Attribute data is stored in dictionaries
            using attribute name keys.
        """
        dInfo = {}
        for categoryName in self.__categoryNameList:
            self.__setCategory(categoryName)
            dI = {}
            dI['DISPLAY_NAME'] = categoryName.lower()
            dI['COLUMN_NAMES'] = self.__currentAttributeNameList
            dI['COLUMN_DISPLAY_NAMES'] = self.__currentAttributeNameList
            dI['PRIMARY_KEYS'] = self.__getKeyAttributeList()
            dI['MANDATORY_COLUMNS'] = self.__getMandatoryAttributeList()
            dI['MANDATORY_COLUMNS_ALT'] = self.__getAltMandatoryAttributeList()
            dI['COLUMN_DISPLAY_ORDER'] = self.__getOrderAttributeList(keyList=dI['PRIMARY_KEYS'], mandatoryList=dI['MANDATORY_COLUMNS'])
            dI['COLUMN_TYPES'] = self.__getTypeD()
            dI['COLUMN_TYPES_ALT'] = self.__getAltTypeD()
            dI['COLUMN_ENUMS'] = self.__getEnumD()
            dI['COLUMN_ENUMS_ALT'] = self.__getAltEnumD()
            dI['COLUMN_DESCRIPTIONS'] = self.__getDescriptionD()
            dI['COLUMN_DESCRIPTIONS_ALT'] = self.__getAltDescriptionD()
            dI['COLUMN_EXAMPLES'] = self.__getExampleD()
            dI['COLUMN_EXAMPLES_ALT'] = self.__getAltExampleD()
            dI['COLUMN_REGEX'] = self.__getRegexD()
            dI['COLUMN_REGEX_ALT'] = self.__getAltRegexD()
            dI['COLUMN_BOUNDARY_VALUES'] = self.__getBoundaryD()
            dI['COLUMN_BOUNDARY_VALUES_ALT'] = self.__getAltBoundaryD()
            dI['COLUMN_DEFAULT_VALUES'] = self.__getDefaultD()
            #
            if (self.__debug):
                logger.debug("PdbxDictionaryInfo(assemble): category %s attributes %r\n" % (categoryName, self.__currentAttributeNameList))
                logger.debug("PdbxDictionaryInfo(assemble): category %s key index list %r\n" % (categoryName, dI['PRIMARY_KEYS']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s mandatory index list %r\n" % (categoryName, dI['MANDATORY_COLUMNS']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s mandatory index list (alt) %r\n" % (categoryName, dI['MANDATORY_COLUMNS_ALT']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s order index list %r\n" % (categoryName, dI['COLUMN_DISPLAY_ORDER']))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type index dictionary %r\n" % (categoryName, dI['COLUMN_TYPES'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type index dictionary (alt) %r\n" % (categoryName, dI['COLUMN_TYPES_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type enum dictionary %r\n" % (categoryName, dI['COLUMN_ENUMS'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type enum dictionary (alt)  %r\n" % (categoryName, dI['COLUMN_ENUMS_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type description dictionary %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type description dictionary (alt) %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type example dictionary %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type example dictionary (alt) %r\n" % (categoryName, dI['COLUMN_DESCRIPTIONS_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type regex dictionary %r\n" % (categoryName, dI['COLUMN_REGEX'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s type regex dictionary (alt) %r\n" % (categoryName, dI['COLUMN_REGEX_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s boundary dictionary %r\n" % (categoryName, dI['COLUMN_BOUNDARY_VALUES'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s boundary dictionary (alt)  %r\n" % (categoryName, dI['COLUMN_BOUNDARY_VALUES_ALT'].items()))
                logger.debug("PdbxDictionaryInfo(assemble): category %s default value dictionary %r\n" % (categoryName, dI['COLUMN_BOUNDARY_VALUES'].items()))

            dInfo[categoryName] = dI
        #
        return dInfo

    def __getEnumIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getEnumList(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[self.__getAttributeIndex(attributeName)] = eList
        return rD

    def __getAltEnumIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getEnumListAlt(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[self.__getAttributeIndex(attributeName)] = eList
        return rD

    def __getDescriptionIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            txt = self.__dApi.getDescription(self.__currentCategoryName, attributeName)
            if txt is not None and len(txt) > 0:
                rD[self.__getAttributeIndex(attributeName)] = txt
        return rD

    def __getAltDescriptionIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            txt = self.__dApi.getDescriptionAlt(self.__currentCategoryName, attributeName)
            if txt is not None and len(txt) > 0:
                rD[self.__getAttributeIndex(attributeName)] = txt
        return rD

    def __getExampleIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getExampleList(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[self.__getAttributeIndex(attributeName)] = eList
        return rD

    def __getAltExampleIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getExampleListAlt(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[self.__getAttributeIndex(attributeName)] = eList
        return rD

    def __getRegexIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            typ = self.__dApi.getTypeRegex(self.__currentCategoryName, attributeName)
            if typ is not None and len(typ) > 0:
                rD[self.__getAttributeIndex(attributeName)] = typ
        return rD

    def __getAltRegexIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            typ = self.__dApi.getTypeRegexAlt(self.__currentCategoryName, attributeName)
            if typ is not None and len(typ) > 0:
                rD[self.__getAttributeIndex(attributeName)] = typ
        return rD

    def __getDefaultIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            typ = self.__dApi.getDefaultValue(self.__currentCategoryName, attributeName)
            if typ is not None and len(typ) > 0:
                rD[self.__getAttributeIndex(attributeName)] = typ
        return rD

    def __getBoundaryIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            bList = self.__dApi.getBoundaryList(self.__currentCategoryName, attributeName)
            if len(bList) > 0:
                rD[self.__getAttributeIndex(attributeName)] = bList
        return rD

    def __getAltBoundaryIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            bList = self.__dApi.getBoundaryListAlt(self.__currentCategoryName, attributeName)
            if len(bList) > 0:
                rD[self.__getAttributeIndex(attributeName)] = bList
        return rD

    def __getTypeIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            dType = self.__dApi.getTypeCode(self.__currentCategoryName, attributeName)
            pType = self.__dApi.getTypePrimitive(self.__currentCategoryName, attributeName)
            isEnum = self.__dApi.isEnumerated(self.__currentCategoryName, attributeName)
            #
            pL = self.__dApi.getParentList(self.__currentCategoryName, attributeName)
            if (self.__debug):
                logger.debug("PdbxDictionaryInfo(assemble): category %s attribute %s parent list %r\n" % (self.__currentCategoryName, attributeName, pL))
            #
            if dType is None:
                (iG, tC, tA) = self.__getUltimateParent(self.__currentCategoryName, attributeName)
                if iG > 0:
                    dType = self.__dApi.getTypeCode(tC, tA)

            if dType is None:
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s missing data type\n" % (self.__currentCategoryName, attributeName))
                dType = 'code'

            if isEnum:
                typ = 'select'
            elif dType in ['float', 'float-range']:
                typ = 'float'
            elif dType in ['int', 'int-range', 'positive_int', 'non_negative_int']:
                typ = 'int'
            elif dType in ['text']:
                typ = 'text'
            elif dType in ['line', 'uline']:
                typ = 'line'
            elif dType in ['code', 'ucode', 'name', 'idname', 'atcode']:
                typ = 'word'
            elif dType.startswith('yyyy-mm'):
                typ = 'date-time'
            elif dType in ['fax', 'phone']:
                typ = 'phone-number'
            elif dType in ['email']:
                typ = 'email'
            elif pType in ['char', 'uchar']:
                typ = 'word'
            else:
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s odd type %s\n" % (self.__currentCategoryName, attributeName, dType))

            rD[self.__getAttributeIndex(attributeName)] = typ

        return rD

    #
    def __getAltTypeIndexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            dType = self.__dApi.getTypeCodeAlt(self.__currentCategoryName, attributeName)
            pType = self.__dApi.getTypePrimitive(self.__currentCategoryName, attributeName)
            isEnum = self.__dApi.isEnumerated(self.__currentCategoryName, attributeName)
            #
            pL = self.__dApi.getParentList(self.__currentCategoryName, attributeName)
            if (self.__debug):
                logger.debug("PdbxDictionaryInfo(assemble): category %s attribute %s parent list %r\n" % (self.__currentCategoryName, attributeName, pL))
            #
            if dType is None:
                (iG, tC, tA) = self.__getUltimateParent(self.__currentCategoryName, attributeName)
                if iG > 0:
                    dType = self.__dApi.getTypeCodeAlt(tC, tA)

            if dType is None:
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s missing data type\n" % (self.__currentCategoryName, attributeName))
                dType = 'code'

            if isEnum:
                typ = 'select'
            elif dType in ['float', 'float-range']:
                typ = 'float'
            elif dType in ['int', 'int-range']:
                typ = 'int'
            elif dType in ['text']:
                typ = 'text'
            elif dType in ['line', 'uline']:
                typ = 'line'
            elif dType in ['code', 'ucode', 'name', 'idname', 'atcode']:
                typ = 'word'
            elif dType.startswith('yyyy-mm'):
                typ = 'date-time'
            elif dType in ['fax', 'phone']:
                typ = 'phone-number'
            elif dType in ['email']:
                typ = 'email'
            elif pType in ['char', 'uchar']:
                typ = 'word'
            else:
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s odd type %s\n" % (self.__currentCategoryName, attributeName, dType))

            rD[self.__getAttributeIndex(attributeName)] = typ

        return rD

    # -----------------------------------
    #
    def __getEnumD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getEnumList(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[attributeName] = eList
        return rD

    def __getAltEnumD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getEnumListAlt(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[attributeName] = eList
        return rD

    def __getDescriptionD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            txt = self.__dApi.getDescription(self.__currentCategoryName, attributeName)
            if txt is not None and len(txt) > 0:
                rD[attributeName] = txt
        return rD

    def __getAltDescriptionD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            txt = self.__dApi.getDescriptionAlt(self.__currentCategoryName, attributeName)
            if txt is not None and len(txt) > 0:
                rD[attributeName] = txt
        return rD

    def __getExampleD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getExampleList(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[attributeName] = eList
        return rD

    def __getAltExampleD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            eList = self.__dApi.getExampleListAlt(self.__currentCategoryName, attributeName)
            if len(eList) > 0:
                rD[attributeName] = eList
        return rD

    def __getRegexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            typ = self.__dApi.getTypeRegex(self.__currentCategoryName, attributeName)
            if typ is not None and len(typ) > 0:
                rD[attributeName] = typ
        return rD

    def __getAltRegexD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            typ = self.__dApi.getTypeRegexAlt(self.__currentCategoryName, attributeName)
            if typ is not None and len(typ) > 0:
                rD[attributeName] = typ
        return rD

    def __getDefaultD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            typ = self.__dApi.getDefaultValue(self.__currentCategoryName, attributeName)
            if typ is not None and len(typ) > 0:
                rD[attributeName] = typ
        return rD

    def __getBoundaryD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            bList = self.__dApi.getBoundaryList(self.__currentCategoryName, attributeName)
            if len(bList) > 0:
                rD[attributeName] = bList
        return rD

    def __getAltBoundaryD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            bList = self.__dApi.getBoundaryListAlt(self.__currentCategoryName, attributeName)
            if len(bList) > 0:
                rD[attributeName] = bList
        return rD

    def __getTypeD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            dType = self.__dApi.getTypeCode(self.__currentCategoryName, attributeName)
            pType = self.__dApi.getTypePrimitive(self.__currentCategoryName, attributeName)
            isEnum = self.__dApi.isEnumerated(self.__currentCategoryName, attributeName)
            #
            pL = self.__dApi.getParentList(self.__currentCategoryName, attributeName)
            if (self.__debug):
                logger.debug("PdbxDictionaryInfo(assemble): category %s attribute %s parent list %r\n" % (self.__currentCategoryName, attributeName, pL))
            #
            if dType is None:
                (iG, tC, tA) = self.__getUltimateParent(self.__currentCategoryName, attributeName)
                if iG > 0:
                    dType = self.__dApi.getTypeCode(tC, tA)

            if dType is None:
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s missing data type\n" % (self.__currentCategoryName, attributeName))
                dType = 'code'

            if isEnum:
                typ = 'select'
            elif dType in ['float', 'float-range']:
                typ = 'float'
            elif dType in ['int', 'int-range', 'positive_int', 'non_negative_int']:
                typ = 'int'
            elif dType in ['text']:
                typ = 'text'
            elif dType in ['line', 'uline']:
                typ = 'line'
            elif dType in ['code', 'ucode', 'name', 'idname', 'atcode', 'emd_id', 'pdb_id']:
                typ = 'word'
            elif dType.startswith('yyyy-mm'):
                typ = 'date-time'
            elif dType in ['fax', 'phone']:
                typ = 'phone-number'
            elif dType in ['email']:
                typ = 'email'
            elif pType in ['char', 'uchar']:
                typ = 'word'
            else:
                typ = 'text'
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s odd type %s\n" % (self.__currentCategoryName, attributeName, dType))

            rD[attributeName] = typ

        return rD

    def __getAltTypeD(self):
        rD = {}
        for attributeName in self.__currentAttributeNameList:
            dType = self.__dApi.getTypeCodeAlt(self.__currentCategoryName, attributeName)
            pType = self.__dApi.getTypePrimitive(self.__currentCategoryName, attributeName)
            isEnum = self.__dApi.isEnumerated(self.__currentCategoryName, attributeName)
            #
            pL = self.__dApi.getParentList(self.__currentCategoryName, attributeName)
            if (self.__debug):
                logger.debug("PdbxDictionaryInfo(assemble): category %s attribute %s parent list %r\n" % (self.__currentCategoryName, attributeName, pL))
            #
            if dType is None:
                (iG, tC, tA) = self.__getUltimateParent(self.__currentCategoryName, attributeName)
                if iG > 0:
                    dType = self.__dApi.getTypeCodeAlt(tC, tA)

            if dType is None:
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s missing data type\n" % (self.__currentCategoryName, attributeName))
                dType = 'code'

            if isEnum:
                typ = 'select'
            elif dType in ['float', 'float-range']:
                typ = 'float'
            elif dType in ['int', 'int-range', 'positive_int', 'non_negative_int']:
                typ = 'int'
            elif dType in ['text']:
                typ = 'text'
            elif dType in ['line', 'uline']:
                typ = 'line'
            elif dType in ['code', 'ucode', 'name', 'idname', 'atcode']:
                typ = 'word'
            elif dType.startswith('yyyy-mm'):
                typ = 'date-time'
            elif dType in ['fax', 'phone']:
                typ = 'phone-number'
            elif dType in ['email']:
                typ = 'email'
            elif pType in ['char', 'uchar']:
                typ = 'word'
            else:
                typ = 'text'
                logger.info("PdbxDictionaryInfo(assemble): category %s attribute %s odd type %s\n" % (self.__currentCategoryName, attributeName, dType))

            rD[attributeName] = typ

        return rD

    def __getOrderIndexList(self, keyList=None, mandatoryList=None):
        iL = []
        if keyList is not None:
            for ii in keyList:
                if not ii in iL:
                    iL.append(ii)
        #
        if mandatoryList is not None:
            for ii in mandatoryList:
                if ii not in iL:
                    iL.append(ii)
        #
        for attrib, ii in self.__currentAttributeD.items():
            if not ii in iL:
                iL.append(ii)
        #
        return iL

    def __getOrderAttributeList(self, keyList=None, mandatoryList=None):
        iL = []
        if keyList is not None:
            for att in keyList:
                if not att in iL:
                    iL.append(att)
        #
        if mandatoryList is not None:
            for att in mandatoryList:
                if att not in iL:
                    iL.append(att)
        #
        for att in self.__currentAttributeNameList:
            if not att in iL:
                iL.append(att)
        #
        return iL

    def __getKeyIndexList(self):
        rL = []
        kyL = self.__dApi.getCategoryKeyList(self.__currentCategoryName)
        for ky in kyL:
            rL.append(self.__getItemIndex(ky))
        return rL

    def __getMandatoryIndexList(self):
        rL = []
        for attributeName in self.__currentAttributeNameList:
            val = self.__dApi.getMandatoryCode(self.__currentCategoryName, attributeName)
            if val is not None and val.upper() in ['Y', 'YES']:
                rL.append(self.__getAttributeIndex(attributeName))
        return rL

    def __getAltMandatoryIndexList(self):
        rL = []
        for attributeName in self.__currentAttributeNameList:
            val = self.__dApi.getMandatoryCodeAlt(self.__currentCategoryName, attributeName)
            if val is not None and val.upper() in ['Y', 'YES']:
                rL.append(self.__getAttributeIndex(attributeName))
        return rL

    def __getKeyAttributeList(self):
        rL = []
        kyL = self.__dApi.getCategoryKeyList(self.__currentCategoryName)
        for ky in kyL:
            rL.append(self.__attributePart(ky))
        return rL

    def __getMandatoryAttributeList(self):
        rL = []
        for attributeName in self.__currentAttributeNameList:
            val = self.__dApi.getMandatoryCode(self.__currentCategoryName, attributeName)
            if val is not None and val.upper() in ['Y', 'YES']:
                rL.append(attributeName)
        return rL

    def __getAltMandatoryAttributeList(self):
        rL = []
        for attributeName in self.__currentAttributeNameList:
            val = self.__dApi.getMandatoryCodeAlt(self.__currentCategoryName, attributeName)
            if val is not None and val.upper() in ['Y', 'YES']:
                rL.append(attributeName)
        return rL

    def __setCategory(self, categoryName):
        """  Set the current category.
        """
        self.__currentCategoryName = categoryName
        self.__currentAttributeNameList = []
        self.__currentAttributeD = {}
        try:
            self.__currentAttributeNameList = sorted(list(set(self.__categoryIndexD[categoryName])))
            for idx, attributeName in enumerate(self.__currentAttributeNameList):
                if attributeName not in self.__currentAttributeD:
                    self.__currentAttributeD[attributeName] = idx
        except Exception as e:
            pass

    def __getParentList(self, itemName):
        try:
            return self.__parentD[itemName]
        except Exception as e:
            return []

    def __getUltimateParent(self, category, attribute):
        """  Return the item(s) at the root of the parent-child tree.
        """
        ok = True
        itemName = "_" + category + "." + attribute
        iGen = 0
        while ok:
            pL = self.__getParentList(itemName)
            #logger.debug("PdbxDictionaryInfo(getUltimateParent): item %s  parent list %r\n" % (itemName,pL))
            if (len(pL) > 0):
                iGen += 1
                itemName = pL[0]
            else:
                ok = False

        return (iGen, self.__categoryPart(itemName), self.__attributePart(itemName))

    def __getAttributeIndex(self, attributeName):
        try:
            return self.__currentAttributeD[attributeName]
        except Exception as e:
            return -1

    def __getItemIndex(self, itemName):
        try:
            return self.__currentAttributeD[self.__attributePart(itemName)]
        except Exception as e:
            return -1

    def __attributePart(self, name):
        i = name.find(".")
        if i == -1:
            return None
        else:
            return name[i + 1:]

    def __categoryPart(self, name):
        tname = ""
        if name.startswith("_"):
            tname = name[1:]
        else:
            tname = name

        i = tname.find(".")
        if i == -1:
            return tname
        else:
            return tname[:i]


class PdbxDictionaryInfoStore(object):
    """ Manage the persistent store of repackaged dictionary and view metadata.

    """

    def __init__(self, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = True
        self.__lfh = log
        #
        # Parameters to tune lock file management --
        self.__timeoutSeconds = 10
        self.__retrySeconds = 0.2
        # placeholder for LockFile object for open/close methods
        self.__lockObj = None
        self.__isExpired = True
        #
        self.__db = None
        #

    def open(self, dbFileName="my.db", flag='r'):
        """ Open the persistent store with the input database file name and access mode.

            flag = 'r read-only'
                   'c/n new/create'
                   'w read/write'
        """
        try:
            self.__lockObj = LockFile(dbFileName, timeoutSeconds=self.__timeoutSeconds, retrySeconds=self.__retrySeconds,
                                      verbose=self.__verbose, log=self.__lfh)
            self.__lockObj.acquire()
            self.__db = shelve.open(dbFileName, flag=flag, protocol=pickle.HIGHEST_PROTOCOL)
            return True
        except Exception as e:
            if self.__lockObj is not None:
                self.__lockObj.release()
            if (self.__verbose):
                logger.debug("+ERROR- PdbxDictInfoStore.open() write failed for file %s\n" % dbFileName)
            if (self.__debug):
                logger.exception("Failing with %s" % str(e))
            return False

    def close(self):
        """ Close the persistent store
        """
        try:
            self.__db.shelve.close()
            if self.__lockObj is not None:
                self.__lockObj.release()
            return True
        except Exception as e:
            if self.__lockObj is not None:
                self.__lockObj.release()
            return False

    def moveStore(self, srcDbFilePath, dstDbFilePath):
        """ Move source store to destination store with destination locking.

            No locking is performed on the source file.
        """
        with LockFile(dstDbFilePath, timeoutSeconds=self.__timeoutSeconds, retrySeconds=self.__retrySeconds, verbose=self.__verbose, log=self.__lfh) as lf:
            retVal = self.__moveStore(srcDbFilePath, dstDbFilePath)
        return retVal

    def store(self, dbFileName="my.db", od=None, ov=None):
        """ Store the input object dictionary and optional view object in persistent database
        """
        with LockFile(dbFileName, timeoutSeconds=self.__timeoutSeconds, retrySeconds=self.__retrySeconds, verbose=self.__verbose, log=self.__lfh) as lf:
            retVal = self.__storeShelve(dbFileName, od=od, ov=ov)
        return retVal

    def getIndex(self, dbFileName="my.db"):
        """ Recover the index of the persistent store.
        """
        with LockFile(dbFileName, timeoutSeconds=self.__timeoutSeconds, retrySeconds=self.__retrySeconds, verbose=self.__verbose, log=self.__lfh) as lf:
            retVal = self.__indexShelve(dbFileName)
        return retVal

    def fetchOneObject(self, dbFileName="my.db", objectName=None):
        """  Fetch a single object from a named container.  This is atomic operation with respect to the store.
        """
        with LockFile(dbFileName, timeoutSeconds=self.__timeoutSeconds, retrySeconds=self.__retrySeconds, verbose=self.__verbose, log=self.__lfh) as lf:
            retVal = self.__fetchOneObjectShelve(dbFileName, objectName)
        return retVal

    def fetchViewObject(self, dbFileName="my.db"):
        """  Fetch a single object from a named container.  This is atomic operation with respect to the store.
        """
        with LockFile(dbFileName, timeoutSeconds=self.__timeoutSeconds, retrySeconds=self.__retrySeconds, verbose=self.__verbose, log=self.__lfh) as lf:
            retVal = self.__fetchOneObjectShelve(dbFileName, objectName="__view__")
        return retVal

    def fetchObject(self, objectName=None):
        """ Fetch object from container from an open store.

            Use this method to extract multiple objects from a store.
        """
        return self.__fetchObjectShelve(objectName)

    def __moveStore(self, src, dst):
        """  Internal method to perform file move.
        """
        try:
            shutil.move(src, dst)
            return True
        except Exception as e:
            if (self.__verbose):
                logger.error("+ERROR- PdbxDictInfoStore.__moveStore() move failed for file %s\n" % src)
            if (self.__debug):
                logger.exception("Failing with %s" % str(e))
            return False

    def __storeShelve(self, dbFileName="my.db", od=None, ov=None):
        """  Create a new persistent store using the internal object dictionary and optional view object dictionary.
        """
        if od is None:
            return False

        try:
            db = shelve.open(dbFileName, flag='c', protocol=pickle.HIGHEST_PROTOCOL)
            for k, v in od.items():
                db[k] = od[k]
            db['__index__'] = list(od.keys())
            db['__view__'] = ov
            db.close()
            return True
        except Exception as e:
            if (self.__verbose):
                logger.error("+ERROR- PdbxDictInfoStore.__storeShelve() shelve store failed for file %s\n" % dbFileName)
            if (self.__debug):
                logger.exception("Failing with %s" % str(e))
            return False

    def __indexShelve(self, dbFileName="my.db"):
        """  Recover the index/list of objects in the persistent store.
        """
        try:
            indexL = []
            db = shelve.open(dbFileName, flag='r', protocol=pickle.HIGHEST_PROTOCOL)
            indexL = db['__index__']
            db.close()
            return indexL
        except Exception as e:
            if (self.__verbose):
                logger.error("+ERROR- PdbxDictInfoStore.__indexShelve() shelve index failed for file %s\n" % dbFileName)
            if (self.__debug):
                logger.exception("Failing with %s" % str(e))
            return {}

    def __fetchOneObjectShelve(self, dbFileName="my.db", objectName=None):
        """ Recover the object from from persistent store corresponding to the
            object name.
        """
        try:
            #
            db = shelve.open(dbFileName, flag='r', protocol=pickle.HIGHEST_PROTOCOL)
            d = db[objectName]
            db.close()
            return d
        except Exception as e:
            if (self.__verbose):
                logger.error("+ERROR- PdbxDictInfoStore.__fetchOneObjectShelve() shelve fetch failed for file %s %s\n"
                             % (dbFileName, objectName))
            if (self.__debug):
                logger.exception("Failing with %s" % str(e))
            return None

    def __fetchObjectShelve(self, objectName=None):
        """ Recover the object from from persistent store corresponding to the
            input object name.

            shelve store must be opened from prior call.
        """
        try:
            d = self.__db[objectName]
            return d
        except Exception as e:
            if (self.__debug):
                logger.error("+ERROR- PdbxDictInfoStore.__fetchObjectShelve() shelve fetch failed %s\n"
                             % (objectName))
            if (self.__debug):
                logger.exception("Failing with %s" % str(e))
            #
            return None



##
# File: LockFile.py
# Date: 21-Feb-2012  John Westbrook
#
# Update:
#
##
"""
Class implementing a cross-platform file locking strategy using an auxiliary lock file.

"""

import os
import sys
import time
import errno


class LockFileTimeoutException(Exception):
    pass


class LockFile(object):
    """ A simple cross-platform file locking utilitiy class using an auxiliary
        lock file.

        This class provides internal methods supporting context-management
        (e.g. with statement support).  For example,

        with LockFile("target-file.db", timeoutSeconds=2) as lock:
            # - process/update the target-file.db.

        # - At the end of this 'with' clause the lock is automatically removed.

    """

    def __init__(self, filePath, timeoutSeconds=15, retrySeconds=.2, verbose=False, log=sys.stderr):
        """ Prepare the file locker. Specify the file to lock and optionally
            the maximum timeoutSeconds and the retrySeconds between each attempt to lock.

            It is assumed that the locking file will be created within the
            path of the target file.
        """
        self.__isLocked = False
        self.__lockFilePath = os.path.join(filePath + ".lock")
        self.__filePath = filePath
        #
        self.__timeoutSeconds = timeoutSeconds
        self.__retrySeconds = retrySeconds
        #
        self.__lfh = log
        self.__verbose = verbose
        self.__debug = False
        #

    def acquire(self):
        """ Create the lock if no lock file exists.  If a lockfile exists then
            repeat the test every retrySeconds.

            If the lock cannot be acquired within  'timeoutSeconds' then
            throw an exception.

        """
        timeBegin = time.time()
        while True:
            try:
                self.__fd = os.open(self.__lockFilePath, os.O_CREAT | os.O_EXCL | os.O_RDWR)
                if (self.__debug):
                    logger.info("LockFile(acquire) created lock file %s\n" % self.__lockFilePath)
                break
            except OSError as myErr:
                if myErr.errno != errno.EEXIST:
                    # pass on some unanticipated problem ---
                    raise
                # handle timeout and retry -
                if (time.time() - timeBegin) >= self.__timeoutSeconds:
                    raise LockFileTimeoutException("LockFile(acquire) Internal timeout of %d (seconds) exceeded for %s" %
                                                   (self.__timeoutSeconds, self.__filePath))
                time.sleep(self.__retrySeconds)
        #
        self.__isLocked = True

    def release(self):
        """ Remove an existing lock file.
        """
        if self.__isLocked:
            os.close(self.__fd)
            os.unlink(self.__lockFilePath)
            self.__isLocked = False
            if (self.__debug):
                logger.info("LockFile(release) removed lock file %s\n" % self.__lockFilePath)

    def __enter__(self):
        """ Internal method for Context-management support.  Invoked at the beginning of a 'with' clause.
        """
        if not self.__isLocked:
            self.acquire()
        return self

    def __exit__(self, type, value, traceback):
        """ Internal method for Context-management support.  Invoked at the end of a 'with' clause.
        """
        if self.__isLocked:
            self.release()

    def __del__(self):
        """ Internal method to cleanup any lingering lock file.
        """
        self.release()

if __name__ == "__main__":
    pass



##
# File: PdbxPrd.py
# Date: 28-Nov-2012
#
# Update:
#  05-May-2013  jdw Add class PdbxReferenceMoleculeList()
##
"""
A collection of accessor classes supporting PRD & Family dictionary data categories.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.01"

import sys
import os

import logging
logger = logging.getLogger(__name__)


class PdbxReferenceMolecule(object):
    ''' Accessor methods for PRD reference molecule top-level attributes.

    '''

    def __init__(self, rowD=None, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def get(self):
        return self.__atD

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getChemCompId(self):
        return self.__getAttribute('chem_comp_id')

    def setChemCompId(self, value):
        return self.__setAttribute('chem_comp_id', value)

    def getName(self):
        return self.__getAttribute('name')

    def setName(self, value):
        return self.__setAttribute('name', value)

    def getRepresentAs(self):
        return self.__getAttribute('represent_as')

    def setRepresentAs(self, value):
        return self.__setAttribute('represent_as', value)

    def getFormula(self):
        return self.__getAttribute('formula')

    def setFormula(self, value):
        return self.__setAttribute('formula', value)

    def getFormulaWeight(self):
        return self.__getAttribute('formula_weight')

    def setFormulaWeight(self, value):
        return self.__setAttribute('formula_weight', value)

    def getType(self):
        return self.__getAttribute('type')

    def setType(self, value):
        return self.__setAttribute('type', value)

    def getTypeEvidence(self):
        return self.__getAttribute('type_evidence_code')

    def setTypeEvidence(self, value):
        return self.__setAttribute('type_evidence_code', value)

    def getClass(self):
        return self.__getAttribute('class')

    def setClass(self, value):
        return self.__setAttribute('class', value)

    def getClassEvidence(self):
        return self.__getAttribute('class_evidence_code')

    def setClassEvidence(self, value):
        return self.__setAttribute('class_evidence_code', value)

    def getReleaseStatus(self):
        return self.__getAttribute('release_status')

    def setReleaseStatus(self, value):
        return self.__setAttribute('release_status', value)

    def getReplacedBy(self):
        return self.__getAttribute('replaced_by')

    def setReplacedBy(self, value):
        return self.__setAttribute('replaced_by', value)

    def getReplaces(self):
        return self.__getAttribute('replaces')

    def setReplaces(self, value):
        return self.__setAttribute('replaces', value)

    def getDescription(self):
        return self.__getAttribute('description')

    def setDescription(self, value):
        return self.__setAttribute('description', value)

    def getCompoundDetails(self):
        return self.__getAttribute('compound_details')

    def setCompoundDetails(self, value):
        return self.__setAttribute('compound_details', value)

    def getRepresentativePdbIdCode(self):
        return self.__getAttribute('representative_PDB_id_code')

    def setRepresentativePdbIdCode(self, value):
        return self.__setAttribute('representative_PDB_id_code', value)


class PdbxReferenceMoleculeList(object):
    ''' Accessor methods for Family PRD membership records

    '''

    def __init__(self, rowD=None, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def get(self):
        return self.__atD

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getPrdId(self):
        return self.__getAttribute('prd_id')

    def setPrdId(self, value):
        return self.__setAttribute('prd_id', value)

    def getFamilyPrdId(self):
        return self.__getAttribute('family_prd_id')

    def setFamilyPrdId(self, value):
        return self.__setAttribute('family_prd_id', value)

    def new(self):
        self.setPrdId(None)
        self.setFamilyPrdId(None)
        return self.__atD


class PdbxPrdAudit(object):
    ''' Accessor methods for PRD definition audit records.

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def new(self):
        self.setId(None)
        self.setActionType(None)
        self.setDate(None)
        self.setAnnotator(None)
        self.setProcessingSite(None)
        self.setDetails(None)
        return self.__atD

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getActionType(self):
        return self.__getAttribute('action_type')

    def setActionType(self, value):
        return self.__setAttribute('action_type', value)

    def getProcessingSite(self):
        return self.__getAttribute('processing_site')

    def setProcessingSite(self, value):
        return self.__setAttribute('processing_site', value)

    def getDate(self):
        return self.__getAttribute('date')

    def setDate(self, value):
        return self.__setAttribute('date', value)

    def getAnnotator(self):
        return self.__getAttribute('annotator')

    def setAnnotator(self, value):
        return self.__setAttribute('annotator', value)

    def getDetails(self):
        return self.__getAttribute('details')

    def setDetails(self, value):
        return self.__setAttribute('details', value)


class PdbxReferenceEntityPolySeq(object):
    ''' Accessor methods PRD reference polymer sequences

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getReferenceEntityId(self):
        return self.__getAttribute('ref_entity_id')

    def setReferenceEntityId(self, value):
        return self.__setAttribute('ref_entity_id', value)

    def getSequenceNumber(self):
        return self.__getAttribute('num')

    def setSequenceNumber(self, value):
        return self.__setAttribute('num', value)

    def getComponentId(self):
        return self.__getAttribute('mon_id')

    def setComponentId(self, value):
        return self.__setAttribute('mon_id', value)

    def getParentComponentId(self):
        return self.__getAttribute('parent_mon_id')

    def setParentComponentId(self, value):
        return self.__setAttribute('parent_mon_id', value)

    def getHeterogeneityFlag(self):
        return self.__getAttribute('hetero')

    def setHeterogeneityFlag(self, value):
        return self.__setAttribute('hetero', value)

    def getObservedFlag(self):
        return self.__getAttribute('observed')

    def setObservedFlag(self, value):
        return self.__setAttribute('observed', value)


class PdbxReferenceEntityPolyLink(object):
    ''' Accessor methods PRD reference polymer linkages

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getReferenceEntityId(self):
        return self.__getAttribute('ref_entity_id')

    def setReferenceEntityId(self, value):
        return self.__setAttribute('ref_entity_id', value)

    def getSequenceNumber1(self):
        return self.__getAttribute('entity_seq_num_1')

    def setSequenceNumber1(self, value):
        return self.__setAttribute('entity_seq_num_1', value)

    def getSequenceNumber2(self):
        return self.__getAttribute('entity_seq_num_2')

    def setSequenceNumber1(self, value):
        return self.__setAttribute('entity_seq_num_2', value)

    def getComponentId1(self):
        return self.__getAttribute('comp_id_1')

    def setComponentId1(self, value):
        return self.__setAttribute('comp_id_1', value)

    def getComponentId2(self):
        return self.__getAttribute('comp_id_2')

    def setComponentId2(self, value):
        return self.__setAttribute('comp_id_2', value)

    def getAtomId1(self):
        return self.__getAttribute('atom_id_1')

    def setAtomId1(self, value):
        return self.__setAttribute('atom_id_1', value)

    def getAtomId2(self):
        return self.__getAttribute('atom_id_2')

    def setAtomId2(self, value):
        return self.__setAttribute('atom_id_2', value)

    def getValueOrder(self):
        return self.__getAttribute('value_order')

    def setValueOrder(self, value):
        return self.__setAttribute('value_order', value)

    def getComponentId(self):
        return self.__getAttribute('component_id')

    def setComponentId(self, value):
        return self.__setAttribute('component_id', value)


class PdbxReferenceEntityNonPoly(object):
    ''' Accessor methods PRD reference non-polymer entities

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getReferenceEntityId(self):
        return self.__getAttribute('ref_entity_id')

    def setReferenceEntityId(self, value):
        return self.__setAttribute('ref_entity_id', value)

    def getComponentId(self):
        return self.__getAttribute('comp_id')

    def setComponentId(self, value):
        return self.__setAttribute('comp_id', value)

    def getType(self):
        return self.__getAttribute('type')

    def setType(self, value):
        return self.__setAttribute('type', value)

    def getDetails(self):
        return self.__getAttribute('details')

    def setDetails(self, value):
        return self.__setAttribute('details', value)


class PdbxReferenceEntityLink(object):
    ''' Accessor methods PRD reference entity linkages

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getLinkId(self):
        return self.__getAttribute('link_id')

    def setLinkId(self, value):
        return self.__setAttribute('link_id', value)

    def getLinkClass(self):
        return self.__getAttribute('link_class')

    def setLinkClass(self, value):
        return self.__setAttribute('link_class', value)

    def getReferenceEntityId1(self):
        return self.__getAttribute('ref_entity_id_1')

    def setReferenceEntityId1(self, value):
        return self.__setAttribute('ref_entity_id_1', value)

    def getReferenceEntityId2(self):
        return self.__getAttribute('ref_entity_id_2')

    def setReferenceEntityId2(self, value):
        return self.__setAttribute('ref_entity_id_2', value)

    def getSequenceNumber1(self):
        return self.__getAttribute('entity_seq_num_1')

    def setSequenceNumber1(self, value):
        return self.__setAttribute('entity_seq_num_1', value)

    def getSequenceNumber2(self):
        return self.__getAttribute('entity_seq_num_2')

    def setSequenceNumber1(self, value):
        return self.__setAttribute('entity_seq_num_2', value)

    def getComponentId1(self):
        return self.__getAttribute('comp_id_1')

    def setComponentId1(self, value):
        return self.__setAttribute('comp_id_1', value)

    def getComponentId2(self):
        return self.__getAttribute('comp_id_2')

    def setComponentId2(self, value):
        return self.__setAttribute('comp_id_2', value)

    def getAtomId1(self):
        return self.__getAttribute('atom_id_1')

    def setAtomId1(self, value):
        return self.__setAttribute('atom_id_1', value)

    def getAtomId2(self):
        return self.__getAttribute('atom_id_2')

    def setAtomId2(self, value):
        return self.__setAttribute('atom_id_2', value)

    def getValueOrder(self):
        return self.__getAttribute('value_order')

    def setValueOrder(self, value):
        return self.__setAttribute('value_order', value)

    def getComponentId1(self):
        return self.__getAttribute('component_id_1')

    def setComponentId1(self, value):
        return self.__setAttribute('component_id_1', value)

    def getComponentId2(self):
        return self.__getAttribute('component_id_2')

    def setComponentId2(self, value):
        return self.__setAttribute('component_id_2', value)


class PdbxReferenceEntityList(object):
    ''' Accessor methods PRD reference entities

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getReferenceEntityId(self):
        return self.__getAttribute('ref_entity_id')

    def setReferenceEntityId(self, value):
        return self.__setAttribute('ref_entity_id', value)

    def getComponentId(self):
        return self.__getAttribute('component_id')

    def setComponentId(self, value):
        return self.__setAttribute('component_id', value)

    def getType(self):
        return self.__getAttribute('type')

    def setType(self, value):
        return self.__setAttribute('type', value)

    def getDetails(self):
        return self.__getAttribute('details')

    def setDetails(self, value):
        return self.__setAttribute('details', value)


class PdbxReferenceEntityPoly(object):
    ''' Accessor methods PRD reference polymer entities

    '''

    def __init__(self, rowD, verbose=True, log=sys.stderr):
        self.__verbose = verbose
        self.__debug = False
        self.__lfh = log
        self.__atD = rowD if rowD is not None else {}

    def __getAttribute(self, name):
        try:
            return self.__atD[name]
        except Exception as e:
            return None

    def __setAttribute(self, name, value):
        try:
            self.__atD[name] = value
            return True
        except Exception as e:
            return False

    def getId(self):
        return self.__getAttribute('prd_id')

    def setId(self, value):
        return self.__setAttribute('prd_id', value)

    def getReferenceEntityId(self):
        return self.__getAttribute('ref_entity_id')

    def setReferenceEntityId(self, value):
        return self.__setAttribute('ref_entity_id', value)

    def getType(self):
        return self.__getAttribute('type')

    def setType(self, value):
        return self.__setAttribute('type', value)

    def getDatabaseCode(self):
        return self.__getAttribute('db_code')

    def setDatabaseName(self, value):
        return self.__setAttribute('db_name', value)



##
# File:    PdbxAnnFeatureTests.py
# Author:  jdw
# Date:    17-Feb-2012
# Version: 0.001
#
# Updated:
#        23-Oct-2012 jdw  Update path and reorganize
##

"""
Unit tests for classes supporting annotation feature data files.

"""
__docformat__ = "restructuredtext en"
__author__    = "John Westbrook"
__email__     = "john.westbrook@rcsb.org"
__license__   = "Apache 2.0"
__version__   = "V0.01"


import sys, unittest, traceback
import sys, time, os, os.path, shutil

from mmcif_utils.pdb.PdbxAnnFeature    import PdbxAnnFeatureCategoryDefinition
from mmcif_utils.pdb.PdbxAnnFeature    import PdbxAnnFeatureReader,PdbxAnnFeatureUpdater
#from mmcif.api.PdbxContainers import *
from mmcif.io.PdbxWriter     import PdbxWriter


class PdbxAnnFeatureTests(unittest.TestCase):
    def setUp(self):
        self.__lfh=sys.stderr
        self.__verbose=True
        self.__pathInputFile       =os.path.join(HERE, "data", "carbohydrate_features_1.cif"
        self.__pathOutputFile      ="testAnnFeatureUpdFile.cif"
        self.__pathOutputFile2     ="testAnnFeatureOutFile.cif"

    def tearDown(self):
        pass

    def testGetCategories(self):
        """Test case -  categories --
        """
        logger.info("\nStarting %s %s\n" % (self.__class__.__name__,
                                               sys._getframe().f_code.co_name))
        try:
            cI=PdbxAnnFeatureCategoryDefinition._categoryInfo
            prd=PdbxAnnFeatureReader(verbose=self.__verbose,log=self.__lfh)
            prd.setFilePath(self.__pathInputFile)
            prd.get()
            catList=prd.getCategoryList()
            logger.info("Category list %r\n" % catList)
            for catName in catList:
                dL=prd.getCategory(catName)
                logger.info("Category %r\n" % dL)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            self.fail()

    def testUpdateReadWrite(self):
        """Test case -  read/write
        """
        logger.info("\nStarting %s %s\n" % (self.__class__.__name__,
                                               sys._getframe().f_code.co_name))
        try:
            cI=PdbxAnnFeatureCategoryDefinition._categoryInfo
            prd=PdbxAnnFeatureUpdater(verbose=self.__verbose,log=self.__lfh)
            prd.read(self.__pathInputFile)
            catList=prd.getCategoryList()
            logger.info("Categories: %r\n" % catList)
            prd.writeFile(filePath=self.__pathOutputFile)
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            self.fail()




def suite():
    return unittest.makeSuite(PdbxAnnFeatureTests,'test')

if __name__ == '__main__':
    unittest.main()



##
# File:    InventoryIoBatchRunner.py
# Author:  jdw
# Date:    15-Aug-2015
# Version: 0.001
#
# Updates:
#   8-Mar-2018 jdw   Use with caution -
##
"""
Batch runner tests for InventoryIo operations ---- Preliminary version --

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"
__version__ = "V0.01"

import sys
import unittest
import time
import os
import shutil
import pickle

import logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s')
logger = logging.getLogger()


from mmcif_utils.inventory.InventoryIo import InventoryIo
from wwpdb.utils.rcsb.MultiProcUtil import MultiProcUtil


class InventoryIoBatchTests(unittest.TestCase):

    def setUp(self):
        self.__lfh = sys.stderr
        self.__verbose = True
        # example files --
        self.__pathIdListFile = "PDBID.LIST"
        self.__overWrite = False

    def tearDown(self):
        pass

    def __getIdList(self, idFilePath):
        idList = []
        ifh = open(idFilePath, 'r')
        for line in ifh:
            if not line.startswith('#'):
                idList.append(line[:-1])
        return idList

    def __getPathList(self, pdbIdList):
        """ return the path list using the input ID list and 2-letter hash path -
        """
        pthL = []
        for pdbId in pdbIdList:
            p = self.__getInputPath(pdbId)
            pthL.append(p)
        return pthL

    def __getInputFilePath(self, pdbId, topPath="/net/to_ftp_arch/ftp-v4.0/pdb/data/structures/divided/structure_factors/"):
        return os.path.join(topPath, pdbId[1:3], 'r' + pdbId + 'sf.ent.gz')

    def __makeOutputFilePath(self, pdbId, topPath="/net/analysis_1/sf-data"):
        pth = os.path.join(topPath, pdbId[1:3], pdbId)
        if not os.access(pth, os.W_OK):
            os.makedirs(pth, 0o775)
        return os.path.join(topPath, pdbId[1:3], pdbId, pdbId + '_sf.cif')

    def __getOutputSummaryFilePath(self, pdbId, topPath="/net/analysis_1/sf-data"):
        return os.path.join(topPath, pdbId[1:3], pdbId, pdbId + '_sf.pic')

    def __exportSummary(self, summaryD, picPath):
        return pickle.dump(summaryD, open(picPath, "wb"))

    def __importSummary(self, picPath):
        if os.access(picPath, os.R_OK):
            return pickle.load(open(picPath, "rb"))
        else:
            return {}

    def __fetch(self, inpPath, outPath):
        try:
            shutil.copyfile(inpPath, outPath)
            return True
        except Exception as e:
            logger.exception("Failing with %s" % str(e))
            return False

    def __copyUnGzip(self, inpFilePath, outFilePath, overWrite=False):
        """
        """
        try:
            if os.access(outFilePath, os.R_OK) and not overWrite:
                return True
            if not os.access(inpFilePath, os.R_OK):
                return False
            cmd = " zcat  %s > %s " % (inpFilePath, outFilePath)
            os.system(cmd)
            return True
        except Exception as e:
            if self.__verbose:
                logger.exception("Failing with %s" % str(e))
            return False

    def getInventoryWorkerMulti(self, dataList, procName, optionsD, workingDir):
        """ Return a list of summary dictionaries --
        """
        retList = []
        oList = []
        try:
            for pdbId in dataList:
                startTime = time.clock()
                logger.info("\nStarting %s %s at %s\n" % (self.__class__.__name__,
                                                          sys._getframe().f_code.co_name,
                                                          time.strftime("%Y %m %d %H:%M:%S", time.localtime())))

                iPth = self.__getInputFilePath(pdbId)
                oPth = self.__makeOutputFilePath(pdbId)
                if self.__copyUnGzip(iPth, oPth, overWrite=self.__overWrite):
                    rD = {'pdb_id': pdbId.lower(), 'summary': {}}
                    #
                    inv = InventoryIo(verbose=self.__verbose, log=self.__lfh)
                    inv.setEnumItemList(['_refln.status', '_refln.pdbx_r_free_flag'])
                    d = inv.getIndex(oPth)
                    sPth = self.__getOutputSummaryFilePath(pdbId)
                    ok = self.__exportSummary(d, sPth)
                    if ok:
                        rD['summary'] = d
                        retList.append(rD)
                        oList.append(pdbId)
                        logger.info("+%s %s return length %d\n" % (self.__class__.__name__, sys._getframe().f_code.co_name, len(d)))
                        if self.__verbose:
                            for k, v in d.items():
                                logger.info("+  %s  %r\n" % (k, v))
                    else:
                        pass
                else:
                    od, of = os.path.split(oPth)
                    shutil.rmtree(od)

                endTime = time.clock()
                logger.info("\nCompleted %s %s at %s (%.2f seconds)\n" % (self.__class__.__name__,
                                                                          sys._getframe().f_code.co_name,
                                                                          time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                                                                          endTime - startTime))

        except Exception as e:
            logger.exception("Failing with %s" % str(e))

        endTime = time.clock()
        logger.info("\nCompleted %s %s at %s (%.2f seconds)\n" % (self.__class__.__name__,
                                                                  sys._getframe().f_code.co_name,
                                                                  time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                                                                  endTime - startTime))

        return oList, retList, []

    def testLoadSummaryMulti(self):
        """Test case - create batch load of sf summary files -
        """
        logger.info("\nStarting %s %s\n" % (self.__class__.__name__, sys._getframe().f_code.co_name))
        startTime = time.time()
        numProc = 8
        try:
            dataList = self.__getIdList(self.__pathIdListFile)
            mpu = MultiProcUtil(verbose=True, log=self.__lfh)
            mpu.set(workerObj=self, workerMethod="getInventoryWorkerMulti")
            ok, failList, retLists, diagList = mpu.runMulti(dataList=dataList, numProc=numProc, numResults=1)
            summaryList = retLists[0]
            endTime0 = time.time()
            logger.info("\nPath list length %d  in %.2f seconds\n" % (len(summaryList), endTime0 - startTime))

        except Exception as e:
            logger.exception("Failing with %s" % str(e))

    def processSummaryTest1(self):
        """ Return a list of summary dictionaries --
        """
        startTime = time.clock()
        logger.info("\nStarting %s %s at %s\n" % (self.__class__.__name__,
                                                  sys._getframe().f_code.co_name,
                                                  time.strftime("%Y %m %d %H:%M:%S", time.localtime())))
        retList = []
        oList = []
        dataList = self.__getIdList(self.__pathIdListFile)
        #
        itemD = {}
        numDataSetD = {}
        numSummaryFiles = 0
        try:
            for pdbId in dataList:
                sPth = self.__getOutputSummaryFilePath(pdbId)
                sD = self.__importSummary(sPth)

                if len(sD) == 0:
                    continue

                if False:
                    logger.info("+%s %s return length %d\n" % (self.__class__.__name__, sys._getframe().f_code.co_name, len(sD)))
                    for k, v in sD.items():
                        logger.info("+  %s  %r\n" % (k, v))
                numSummaryFiles += 1
                #
                # number of data sets -
                #
                numDataSets = sD['CONTAINER_COUNT']
                if numDataSets not in numDataSetD:
                    numDataSetD[numDataSets] = 0
                numDataSetD[numDataSets] += 1
                #
                # Count items -  CONTAINER_INFO  {0: {'OBJ_INFO': {'audit': {'OBJ_LENGTH': 1, 'ITEM_INFO': [('_audit.revision_id', 1), ('_audit.creat
                d = sD['CONTAINER_INFO']
                for iSet, tD in d.items():
                    if iSet != 0:
                        continue
                    objD = tD['OBJ_INFO']
                    for catName, catD in objD.items():
                        for itTup in catD['ITEM_INFO']:
                            if itTup[0] not in itemD:
                                itemD[itTup[0]] = []
                            itemD[itTup[0]].append(pdbId)
            #
            logger.info("Total entries = %d\n" % len(dataList))
            logger.info("Total summary files = %d  %.4f\n" % (numSummaryFiles, float(numSummaryFiles) / len(dataList)))
            #
            totalDataSets = 0
            n = 0
            for k in sorted(numDataSetD.keys()):
                v = numDataSetD[k]
                logger.info(" set count %2d    number %6d   pc %.4f\n" % (k, v, float(v) / float(numSummaryFiles)))
                totalDataSets += (k * v)
                n += v
            #
            logger.info(" Total data set count %7d\n" % totalDataSets)
            logger.info(" Check of the number of summary files %7d\n" % n)

            #
            #  number of sets containing item
            #
            for k in sorted(itemD.keys()):
                vL = itemD[k]
                numSets = len(vL)
                numEntries = len(set(vL))
                logger.info("%-55s   in data sets %6d  (%8.4f) in entries %6d (%8.4f)\n" %
                            (k, numSets, float(numSets) / float(totalDataSets), numEntries, float(numEntries) / float(numSummaryFiles)))

            ok = True
        except Exception as e:
            ok = False
            logger.exception("Failing with %s" % str(e))

        endTime = time.clock()
        logger.info("\nCompleted %s %s at %s (%.2f seconds)\n" % (self.__class__.__name__,
                                                                  sys._getframe().f_code.co_name,
                                                                  time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                                                                  endTime - startTime))

        return ok


def suiteInventoryTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(InventoryIoBatchTests("testLoadSummaryMulti"))


def suiteProcessSummaryTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(InventoryIoBatchTests("processSummaryTest1"))
    return suiteSelect

if __name__ == '__main__':
    if (False):
        mySuite = suiteInventoryTests()
        unittest.TextTestRunner(verbosity=2).run(mySuite)

        mySuite = suiteProcessSummaryTests()
        unittest.TextTestRunner(verbosity=2).run(mySuite)

    mySuite = suiteProcessSummaryTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite)
